/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.admin.osgi.obr;

import java.io.PrintStream;
import java.net.URL;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;

import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.osgi.service.obr.Capability;
import org.osgi.service.obr.Repository;
import org.osgi.service.obr.RepositoryAdmin;
import org.osgi.service.obr.Requirement;
import org.osgi.service.obr.Resolver;
import org.osgi.service.obr.Resource;
import org.ow2.jonas.admin.osgi.StringBufferOutputStream;

public class OBRPlugin {
    private BundleContext m_context = null;

    private static ServiceReference m_repoRef = null;

    private static RepositoryAdmin m_repoAdmin = null;

    private static SimpleTreeNode m_rootNode = null;

    private CreateRootRunnable m_createRootRunnable = new CreateRootRunnable();

    private SetRootRunnable m_setRootRunnable = new SetRootRunnable();

    private static JTree m_bundleTree = null;

    private static HashSet hs;

    public OBRPlugin(final BundleContext context) {
        hs = new HashSet<String>();

        m_context = context;

        synchronized (this) {
            // Listen for registering/unregistering bundle repository services.
            ServiceListener sl = new ServiceListener() {
                public void serviceChanged(ServiceEvent event) {
                    synchronized (OBRPlugin.this) {
                        // Ignore additional services if we already have one.
                        if ((event.getType() == ServiceEvent.REGISTERED) && (m_repoRef != null)) {
                            return;
                        }
                        // Initialize the service if we don't have one.
                        else if ((event.getType() == ServiceEvent.REGISTERED) && (m_repoRef == null)) {
                            lookupService();
                        }
                        // Unget the service if it is unregistering.
                        else if ((event.getType() == ServiceEvent.UNREGISTERING)
                            && event.getServiceReference().equals(m_repoRef)) {
                            m_context.ungetService(m_repoRef);
                            m_repoRef = null;
                            m_repoAdmin = null;
                            // Try to get another service.
                            lookupService();
                        }
                    }
                }
            };

            try {
                m_context.addServiceListener(sl, "(objectClass=" + RepositoryAdmin.class.getName() + ")");
            } catch (InvalidSyntaxException ex) {
                System.err.println("OBRPlugin: " + ex);
            }

            // Now try to manually initialize the shell service
            // since one might already be available.
            lookupService();
        }
    }

    private synchronized void lookupService() {
        if (m_repoAdmin != null) {
            return;
        }
        m_repoRef = m_context.getServiceReference(RepositoryAdmin.class.getName());
        if (m_repoRef == null) {
        } else {
            m_repoAdmin = (RepositoryAdmin) m_context.getService(m_repoRef);
        }

        // Update the model.
        initializeRootNode();
    }

    private void initializeRootNode() {
        synchronized (m_createRootRunnable) {
            new Thread(m_createRootRunnable).start();
        }
    }

    private class CreateRootRunnable implements Runnable {
        public void run() {
            synchronized (OBRPlugin.this) {
                // HACK ALERT: This next if statement is a hack to force
                // the OBR service to retrieve its repository files on
                // this thread, rather than the Swing thread. This hack
                // assumes that this GUI is working with Felix' OBR service,
                // which defers retrieving repository URLs until needed.
                if (m_repoAdmin != null) {
                    m_repoAdmin.listRepositories();
                }

                // Create the new root node and then set it.
                m_rootNode = new SimpleTreeNode(null, m_repoAdmin);
                try {
                    SwingUtilities.invokeAndWait(m_setRootRunnable);
                } catch (Exception ex) {
                    // Ignore.
                }
            }
        }
    }

    private class SetRootRunnable implements Runnable {
        public void run() {
            ((DefaultTreeModel) m_bundleTree.getModel()).setRoot(m_rootNode);
        }
    }

    private static class SimpleTreeNode implements TreeNode {
        private TreeNode m_parent = null;

        private Object m_obj = null;

        private TreeNode[] m_children = null;

        private String m_toString = null;

        public SimpleTreeNode(final TreeNode parent, final Object obj) {
            m_parent = parent;
            m_obj = obj;
        }

        public Object getObject() {
            return m_obj;
        }

        public TreeNode getChildAt(final int index) {
            if (m_children == null) {
                initialize();
            }

            if ((m_children != null) && (index >= 0) && (index < m_children.length)) {
                return m_children[index];
            }

            return null;
        }

        public int getChildCount() {
            if (m_children == null) {
                initialize();
            }
            return (m_children == null) ? 0 : m_children.length;
        }

        public TreeNode getParent() {
            return m_parent;
        }

        public int getIndex(final TreeNode node) {
            if (m_children == null) {
                initialize();
            }
            for (int i = 0; (m_children != null) && (i < m_children.length); i++) {
                if (m_children[i] == node) {
                    return i;
                }
            }
            return -1;
        }

        public boolean getAllowsChildren() {
            return true;
        }

        public boolean isLeaf() {
            return (getChildCount() == 0);
        }

        public Enumeration children() {
            return null;
        }

        private void initialize() {
            // The current node might be the root repository admin.
            if ((m_obj != null) && (m_obj instanceof RepositoryAdmin)) {
                Object[] objs = ((RepositoryAdmin) m_obj).listRepositories();
                if (objs != null) {
                    m_children = new TreeNode[objs.length];
                    for (int i = 0; i < objs.length; i++) {
                        m_children[i] = new SimpleTreeNode(this, objs[i]);
                    }
                }
            } else if (m_obj instanceof Repository) {
                Object[] objs = ((Repository) m_obj).getResources();
                if (objs != null) {
                    m_children = new TreeNode[objs.length];
                    for (int i = 0; i < objs.length; i++) {
                        m_children[i] = new SimpleTreeNode(this, objs[i]);
                    }
                }
            }
        }

        @Override
        public String toString() {
            if (m_toString == null) {
                if (m_obj instanceof RepositoryAdmin) {
                    m_toString = "ROOT";
                } else if (m_obj instanceof Repository) {
                    m_toString = ((Repository) m_obj).getName();
                } else if (m_obj instanceof Resource) {
                    m_toString = ((Resource) m_obj).getPresentationName() + " (" + ((Resource) m_obj).getVersion() + ")";
                } else {
                    m_toString = (m_obj != null) ? m_obj.toString() : "null";
                }
            }
            return m_toString;
        }
    }

    private Object getNodeObject(final String path) {
        String[] steps = path.split("/");
        TreeNode currentnode = m_rootNode;
        for (String step : steps) {
            if (!step.equals("")) {
                int i = Integer.parseInt(step);
                currentnode = currentnode.getChildAt(i);
            }
        }
        return ((SimpleTreeNode) currentnode).getObject();
    }

    private String tree2xml(final SimpleTreeNode node, final String path) {
        String xml = "";
        if (node.getObject() instanceof RepositoryAdmin || node.getObject() instanceof Repository) {
            xml = xml + "<node label=\"" + node.toString() + "\"" + " path=\"" + path + "\"" + ">";
            for (int i = 0; i < node.getChildCount(); i++) {
                xml = xml + tree2xml((SimpleTreeNode) node.getChildAt(i), path + "/" + i);
            }
            xml = xml + "</node>";
        } else if (node.getObject() instanceof Resource) {
            xml = xml + "<node label=\"" + node.toString() + "\"" + " path=\"" + path + "\"" + " />";
        }
        return xml;
    }

    private void printInfo(final PrintStream out, final Object obj) {
        if (obj != null) {
            if (obj instanceof Repository) {
                Repository repo = (Repository) obj;
                out.println(repo.getName());
                out.println("   URL = " + repo.getURL());
                out.println("   Modified = " + new Date(repo.getLastModified()));
            } else if (obj instanceof Resource) {
                Resource res = (Resource) obj;
                out.println(res.getPresentationName());

                // Print properties.
                Map map = res.getProperties();
                Iterator iter = map.entrySet().iterator();
                while (iter.hasNext()) {
                    Map.Entry entry = (Map.Entry) iter.next();
                    out.println("   " + entry.getKey() + " = " + entry.getValue());
                }

                // Print requirements.
                Requirement[] reqs = res.getRequirements();
                for (int i = 0; (reqs != null) && (i < reqs.length); i++) {
                    if (i == 0) {
                        out.println("   requirements:");
                    }
                    out.println("      " + reqs[i].getFilter());
                }

                // Print capabilities.
                Capability[] caps = res.getCapabilities();
                for (int i = 0; (caps != null) && (i < caps.length); i++) {
                    if (i == 0) {
                        out.println("   capabilities:");
                    }
                    out.println("      " + caps[i].getName() + " = " + caps[i].getProperties());
                }
            }
        }
    }

    private void printUnderline(final PrintStream out, final int length) {
        for (int i = 0; i < length; i++) {
            out.print('-');
        }
        out.println("");
    }

    public String create(final BundleContext bc) {
        synchronized (OBRPlugin.this) {
            if (m_repoAdmin == null || m_repoAdmin.listRepositories().length == 0) {
                return null;
            }
            initializeRootNode();
            String xml = tree2xml(m_rootNode, "");
            return xml;
        }
    }

    public String add(final String url, final BundleContext bc) {
        if (!hs.contains(url)) {
            synchronized (OBRPlugin.this) {
                if (m_repoAdmin == null) {
                    return null;
                }

                try {
                    m_repoAdmin.addRepository(new URL(url));
                    hs.add(url);
                } catch (Exception ex) {
                    ex.printStackTrace();
                }

                // Update the table.
                initializeRootNode();
            }
            String xml;
            if (m_rootNode != null) {
                xml = tree2xml(m_rootNode, "");
            } else {
                xml = null;
            }
            return xml;
        }
        return null;
    }

    public String remove(final String paths, final BundleContext bc) {
        synchronized (OBRPlugin.this) {
            String[] pathlist = paths.split("_");
            for (String path : pathlist) {
                if (!path.equals("")) {
                    Object o = getNodeObject(path);
                    if (o instanceof Repository) {
                        m_repoAdmin.removeRepository(((Repository) o).getURL());
                        hs.remove(((Repository) o).getURL().toString());
                    }
                }
            }
            initializeRootNode();
        }
        if (m_repoAdmin == null || m_repoAdmin.listRepositories().length == 0) {
            return null;
        }
        String xml = tree2xml(m_rootNode, "");
        return xml;
    }

    public String refresh(final String paths, final BundleContext bc) {
        synchronized (OBRPlugin.this) {
            String[] pathlist = paths.split("_");
            for (String path : pathlist) {
                if (!path.equals("")) {
                    Object o = getNodeObject(path);
                    if (o instanceof Repository) {
                        try {
                            m_repoAdmin.addRepository(((Repository) o).getURL());
                        } catch (Exception e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }
                    }
                }
            }
            initializeRootNode();
        }
        String xml = tree2xml(m_rootNode, "");
        return xml;
    }

    public String info(final String paths, final BundleContext bc) {
        synchronized (OBRPlugin.this) {
            if (m_repoAdmin == null) {
                return null;
            }
            StringBufferOutputStream sb = new StringBufferOutputStream();
            PrintStream m_out = new PrintStream(sb);
            String[] pathlist = paths.split("_");
            for (int i = 0; i < pathlist.length; i++) {
                if (i != 0) {
                    m_out.println("");
                }
                if (!pathlist[i].equals("")) {
                    Object o = getNodeObject(pathlist[i]);
                    printInfo(m_out, o);
                }
            }
            m_out.println("");
            return sb.toString();
        }
    }

    public String deploystart(final String paths, final BundleContext bc, final boolean start) {
        synchronized (OBRPlugin.this) {
            if (m_repoAdmin == null) {
                return null;
            }
            StringBufferOutputStream sb = new StringBufferOutputStream();
            PrintStream m_out = new PrintStream(sb);
            String[] pathlist = paths.split("_");
            Resolver resolver = m_repoAdmin.resolver();
            for (int i = 0; i < pathlist.length; i++) {
                Object o = getNodeObject(pathlist[i]);
                if (o instanceof Resource) {
                    resolver.add((Resource) o);
                }
            }

            if ((resolver.getAddedResources() != null) && (resolver.getAddedResources().length > 0)) {
                if (resolver.resolve()) {
                    m_out.println("Target resource(s):");
                    printUnderline(m_out, 19);
                    Resource[] resources = resolver.getAddedResources();
                    for (int resIdx = 0; (resources != null) && (resIdx < resources.length); resIdx++) {
                        m_out.println("   " + resources[resIdx].getPresentationName() + " (" + resources[resIdx].getVersion()
                            + ")");
                    }
                    resources = resolver.getRequiredResources();
                    if ((resources != null) && (resources.length > 0)) {
                        m_out.println("\nRequired resource(s):");
                        printUnderline(m_out, 21);
                        for (int resIdx = 0; resIdx < resources.length; resIdx++) {
                            m_out.println("   " + resources[resIdx].getPresentationName() + " ("
                                + resources[resIdx].getVersion() + ")");
                        }
                    }
                    resources = resolver.getOptionalResources();
                    if ((resources != null) && (resources.length > 0)) {
                        m_out.println("\nOptional resource(s):");
                        printUnderline(m_out, 21);
                        for (int resIdx = 0; resIdx < resources.length; resIdx++) {
                            m_out.println("   " + resources[resIdx].getPresentationName() + " ("
                                + resources[resIdx].getVersion() + ")");
                        }
                    }

                    try {
                        m_out.print("\nDeploying...");
                        resolver.deploy(start);
                        m_out.println("done.");
                    } catch (IllegalStateException ex) {
                        m_out.println(ex);
                    }
                } else {
                    Requirement[] reqs = resolver.getUnsatisfiedRequirements();
                    if ((reqs != null) && (reqs.length > 0)) {
                        m_out.println("Unsatisfied requirement(s):");
                        printUnderline(m_out, 27);
                        for (int reqIdx = 0; reqIdx < reqs.length; reqIdx++) {
                            m_out.println("   " + reqs[reqIdx].getFilter());
                            Resource[] resources = resolver.getResources(reqs[reqIdx]);
                            for (int resIdx = 0; resIdx < resources.length; resIdx++) {
                                m_out.println("      " + resources[resIdx].getPresentationName());
                            }
                        }
                    } else {
                        m_out.println("Could not resolve targets.");
                    }
                }

                m_out.println("");
            }
            return sb.toString();
        }
    }
}
